/*
 * File: bind.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */


/*** 

This file is for functions having to do with key bindings, descriptions,
help commands and startup file.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"bind.h"
#include	"buffer.h"
#include	"exec.h"
#include	"fileio.h"
#include	"line.h"
#include	"macros.h"
#include	"messages.h"
#include	"terminal.h"
#include	"utils.h"
#include	"window.h"


/*++

Output a string of output characters

++*/
static void ostring(s)
char *s;                        /* string to output */

{
    if (discmd)
        while (*s)
            mlout(*s++);
}



/*++

Looks up key binding in binding table.  Returns ptr to binding
else ptr to end of binding table. Local helper fn.

++*/
static KEYTAB *get_binding(c)
int c;                          /* key */

{
    KEYTAB *ktp;

    /* scan through binding table for c's entry */
    ktp = &keytab[0];
    while (ktp->k_type != BINDNUL) {
        if (ktp->k_code == c)
            return (ktp);
        ++ktp;
    }

    return (ktp);
}



/*++

Get a command key sequence from the keyboard.
Local helper for functions in bind.c

++*/
static int get_cmd_key(mflag)
int mflag;                      /* meta sequence? */

{
    int c;                      /* character fetched */
    char tok[NSTRING];          /* command incoming */

    /* non-interactive */
    if (clexec) {
        get_macro_line(tok);    /* get the next token */
        return (keystr2keycmd(tok));
    }

    /* interactive */
    if (mflag)
        c = mlgetkey();
    else
        c = get_keystrokes();

    return (c);
}


/*++

Describe the command for a certain key.

++*/
int describe_key()
{
    int c;                      /* key to describe */
    char *ptr;                  /* string pointer to scan output strings */
    char outseq[NSTRING];       /* output buffer for command sequence */

    /* prompt the user to type us a key to describe */
    mlwrite(TEXT13);            /* ": describe-key " */

    /* get cmd and convert to string */
    ostring(cmd2str(c = get_cmd_key(FALSE), &outseq[0]));
    ostring(" ");

    /* find the right function */
    if ((ptr = get_name(exist_binding(c))) == NULL)
        ptr = "Not Bound";

    /* output the command sequence */
    ostring(ptr);
    return (TRUE);
}



/*++

Helper fn for unbind_key

++*/
static int unbind_char(c)
int c;                          /* command key to unbind */

{
    KEYTAB *ktp;                /* pointer into the command table */
    KEYTAB *sktp;               /* saved pointer into the command table */

    /* search the table to see if the key exists */
    ktp = get_binding(c);
    if (ktp->k_type == BINDNUL)
        return (FALSE);

    /* save the pointer and scan to the end of the table */
    sktp = ktp;
    while (ktp->k_type != BINDNUL)
        ++ktp;
    --ktp;                      /* backup to the last legit entry */

    /* copy the last entry to the current one */
    sktp->k_code = ktp->k_code;
    sktp->k_type = ktp->k_type;
    if (sktp->k_type == BINDFNC)
        sktp->k_fp = ktp->k_fp;
    else if (sktp->k_type == BINDBUF)
        sktp->k_buf = ktp->k_buf;

    /* null out the last one */
    ktp->k_code = 0;
    ktp->k_type = BINDNUL;
    ktp->k_fp = NULL;
    ktp->k_buf = NULL;
    return (TRUE);
}


/*++

Clears bindings of unsourced buffers

++*/
int clear_sbinding(name)
char *name;                     /* buffer name */

{
    KEYTAB *ktp;

    /* scan through binding table for name */
    ktp = &keytab[0];
    while (ktp->k_type != BINDNUL) {
        if ((ktp->k_buf != NULL)
            && (strcmp(ktp->k_buf->b_name, name) == 0))
            return (unbind_char(ktp->k_code));
        ++ktp;
    }

    return (FALSE);
}



/*++

Local helper to add new key-binding to end of table

++*/
static int new_binding(ktp, c, kbuf, kfunc)
KEYTAB *ktp;                    /* pointer into the command table */
int c;                          /* command key to bind */
BUFFER *kbuf;                   /* ptr to buffer of element to bind */
int (*kfunc) ();                /* ptr to the requested function to bind to */
{

    /* if we run out of binding room, bitch */
    if (ktp >= &keytab[NBINDS]) {
        mlwrite(TEXT17);        /* "Binding table FULL!" */
        return (FALSE);
    }

    ktp->k_code = c;            /* add keycode */
    if (kfunc != NULL) {        /* and the binding type */
        ktp->k_type = BINDFNC;
        ktp->k_fp = kfunc;      /* and the fn pointer */
    } else {
        ktp->k_type = BINDBUF;
        ktp->k_buf = kbuf;      /* or buffer pointer */
    }
    ++ktp;                      /* make sure the next is null */
    ktp->k_code = 0;
    ktp->k_type = BINDNUL;
    ktp->k_buf = NULL;
    ktp->k_fp = NULL;

    return TRUE;
}



/*++

Add a new key to the key binding table.  This is used by the .rhorc and
by sourced files to rebind existing bindings to a user-defined binding.

++*/
int bind_to_key()
{
    int c;                      /* command key to bind */
    int (*kfunc) ();            /* ptr to the requested function to bind to */
    KEYTAB *ktp;                /* pointer into the command table */
    char outseq[80];            /* output buffer for keystroke sequence */

    /* prompt the user to type in a key to bind */
    /* get the function name to bind it to */
    kfunc = prompt4command(TEXT15);     /* ": bind-to-key " */
    if (kfunc == NULL) {
        mlwrite(TEXT16);        /* "[No such function]" */
        return (FALSE);
    }
    if (clexec == FALSE) {
        ostring(" ");
        TTflush();
    }

    /* get the command sequence to bind */
    c = get_cmd_key((kfunc == meta_prefix) || (kfunc == ctlx) ||
                    (kfunc == universal_argument)
                    || (kfunc == abort_command));

    if (clexec == FALSE) {      /* print cmd string */
        ostring(cmd2str(c, &outseq[0]));
    }

    /* if the function is a unique prefix key */
    if (kfunc == universal_argument || kfunc == abort_command
        || kfunc == quote) {

        /* search for an existing binding for the prefix key */
        ktp = &keytab[0];
        while (ktp->k_type != BINDNUL) {
            if (ktp->k_type == BINDFNC && ktp->k_fp == kfunc)
                unbind_char(ktp->k_code);
            ++ktp;
        }

        /* reset the appropriate global prefix variable */
        if (kfunc == universal_argument)
            reptc = c;
        if (kfunc == abort_command)
            abortc = c;
        if (kfunc == quote)
            quotec = c;
    }

    /* search the table to see if c exists */
    ktp = get_binding(c);

    if (ktp->k_type != BINDNUL) {       /* if it exists, change it */
        ktp->k_fp = kfunc;
        ktp->k_type = BINDFNC;
    } else                      /* otherwise, add it to the end */
        new_binding(ktp, c, NULL, kfunc);

    /* if meta key rebound , make search terminators follow it  */
    if (kfunc == meta_prefix) {
        sterm = c;
        isterm = c;
    }

    return (TRUE);
}


/*++

Bind a key to a macro in the key binding table.  This is used by the
.rhorc and by sourced files to bind a macro in that file to a key.

++*/
int macro_to_key()
{
    int c;                      /* command key to bind */
    BUFFER *kmacro;             /* ptr to buffer of macro to bind to key */
    KEYTAB *ktp;                /* pointer into the command table */
    int status;                 /* error return */
    char outseq[80];            /* output buffer for keystroke sequence */
    char bufn[LENPATH];         /* buffer to hold macro name */
    char *bufn2;

    /* get the buffer name to use */
    if ((status = mlreply(TEXT215, &bufn[1], LENPATH - 2)) != TRUE)
        return (status);        /* ": macro-to-key " */

    /* build the response string for later */
    strcpy(outseq, TEXT215);
    strcat(outseq, &bufn[1]);

    /* translate it to a buffer pointer */
    bufn[0] = '[';
    strcat(bufn, "]");
    bufn2 = strdup(bufn);
    bufn2[strlen(bufn2) - 1] = ')';
    if ((kmacro = exist_buffer(bufn, FALSE, 0)) == NULL) {
        if ((kmacro = exist_buffer(bufn2, FALSE, 0)) == NULL) {
            mlwrite(TEXT130);   /*     "Macro not defined" */
            return (FALSE);
        }
    }

    strcat(outseq, " ");
    mlwrite(outseq);

    /* get the command sequence to bind */
    c = get_cmd_key(FALSE);

    /* print cmd string */
    ostring(cmd2str(c, &outseq[0]));

    /* search the table to see if c exists */
    ktp = get_binding(c);

    if (ktp->k_type != BINDNUL) {       /* if exists, change it */
        ktp->k_buf = kmacro;
        ktp->k_type = BINDBUF;
    } else                      /* else add to end */
        new_binding(ktp, c, kmacro, NULL);

    return (TRUE);
}


/*++

Delete a key from the key binding table.

++*/
int unbind_key()
{
    int c;                      /* command key to unbind */
    char outseq[80];            /* output buffer for keystroke sequence */

    /* prompt the user to type in a key to unbind */
    mlwrite(TEXT18);            /* ": unbind-key " */

    /* get the command sequence to unbind */
    c = get_cmd_key(FALSE);     /* get a command sequence */

    /* change it to something we can print */
    /* and dump it out */
    ostring(cmd2str(c, &outseq[0]));

    /* if it isn't bound, bitch */
    if (unbind_char(c) == FALSE) {
        mlwrite(TEXT19);        /* "[Key not bound]" */
        return (FALSE);
    }
    return (TRUE);
}


/*++

Engine for describe_bindings and apropos.
Builds a list of key-bindings.

++*/
static int build_list(type, mstring)
int type;                       /* true = full list,   false = partial list */
char *mstring;                  /* match string if a partial list */

{
    KEYTAB *ktp;                /* pointer into the command table */
    NBIND *nptr;                /* pointer into the name binding table */
    BUFFER *listbuf;            /* buffer to put binding list into */
    BUFFER *bp;                 /* buffer ptr for function scan */
    int cpos;                   /* current position to use in outseq */
    char outseq[80];            /* output buffer for keystroke sequence */

    /* get a buffer for the binding list */
    listbuf = exist_buffer(TEXT21, TRUE, BFINVS);       /* "Binding list" */
    clear_buffer(listbuf);

    /* build the contents of this window, inserting it line by line */
    nptr = &names[0];
    while (nptr->n_func != NULL) {

        /* add in the command name */
        strcpy(outseq, nptr->n_name);
        cpos = (int) strlen(outseq);

        /* if we are executing an apropos command..... */
        if (type == FALSE &&
            /* and current string doesn't include the search string */
            strstr(outseq, mstring) == FALSE) {
            ++nptr;
            continue;
        }

        /* search down any keys bound to this */
        ktp = &keytab[0];
        while (ktp->k_type != BINDNUL) {
            if (ktp->k_type == BINDFNC && ktp->k_fp == nptr->n_func) {
                /* padd out some spaces */
                while (cpos < 25)
                    outseq[cpos++] = ' ';

                /* add in the command sequence */
                (void) cmd2str(ktp->k_code, &outseq[cpos]);

                /* and add it as a line into the buffer */
                if (addline(listbuf, outseq) != TRUE)
                    return (FALSE);

                cpos = 0;       /* and clear the line */
            }
            ++ktp;
        }

        /* if no key was bound, we need to dump it anyway */
        if (cpos > 0) {
            outseq[cpos] = 0;
            if (addline(listbuf, outseq) != TRUE)
                return (FALSE);
        }

        ++nptr;
    }

    /* add a blank line between the key and macro lists */
    if (addline(listbuf, "") != TRUE)
        return (FALSE);

    /* scan all buffers looking for macroes and their bindings */
    bp = bheadp;
    while (bp) {

        /* is this buffer a macro? */
        if (bp->b_name[0] != '[') {
            bp = bp->b_bufp;
            continue;
        }

        /* add in the command name */
        strcpy(outseq, bp->b_name);
        cpos = (int) strlen(outseq);

        /* if apropos */
        if (type == FALSE &&
            /* and current string doesn't include the search string */
            strstr(outseq, mstring) == FALSE) {
            bp = bp->b_bufp;
            continue;
        }

        /* search down any keys bound to this macro */
        ktp = &keytab[0];
        while (ktp->k_type != BINDNUL) {
            if (ktp->k_type == BINDBUF && ktp->k_buf == bp) {
                /* padd out some spaces */
                while (cpos < 25)
                    outseq[cpos++] = ' ';

                /* add in the command sequence */
                (void) cmd2str(ktp->k_code, &outseq[cpos]);

                /* and add it as a line into the buffer */
                if (addline(listbuf, outseq) != TRUE)
                    return (FALSE);

                cpos = 0;       /* and clear the line */
            }
            ++ktp;
        }

        /* if no key was bound, we need to dump it anyway */
        if (cpos > 0) {
            outseq[cpos] = 0;
            if (addline(listbuf, outseq) != TRUE)
                return (FALSE);
        }

        bp = bp->b_bufp;
    }

    wpopup(listbuf);
    mlerase();                  /* clear the mode line */
    return (TRUE);
}


/*++

Bring up a buffer and list the key bindings into it in view mode.
   
++*/
int describe_bindings()
{
    return (build_list(TRUE, ""));
}



/*++

List functions that match a substring.

++*/
int apropos()
{
    char mstring[NSTRING];      /* string to match cmd names to */
    int status;                 /* status return */

    status = mlreply(TEXT20, mstring, NSTRING - 1);     /* "Apropos string: " */
    if (status != TRUE)
        return (status);

    return (build_list(FALSE, mstring));
}



/*++

Change a key command to a string we can print out.
Return the string passed in.

++*/
char *cmd2str(c, seq)
int c;                          /* sequence to translate */
char *seq;                      /* destination string for sequence */

{
    char *ptr;                  /* pointer into current position in sequence */

    ptr = seq;

    /* apply ^X sequence if needed */
    if (c & CTLX) {
        *ptr++ = '^';
        *ptr++ = 'X';
    }

    /* apply Shifted sequence if needed */
    if (c & SHFT) {
        *ptr++ = 'S';
        *ptr++ = '-';
    }

    /* apply meta sequence if needed */
    if (c & META) {
        *ptr++ = 'M';
        *ptr++ = '-';
    }

    /* apply SPEC sequence if needed */
    if (c & SPEC) {
        *ptr++ = 'F';
        *ptr++ = 'N';
    }

    /* apply control sequence if needed */
    if (c & CTRL) {

        /* non normal spaces look like @ */
        if (ptr == seq && ((c & 255) == ' '))
            c = '@';

        *ptr++ = '^';
    }

    c = c & 255;                /* strip the prefixes */

    /* and output the final sequence */
    *ptr++ = (char) c;
    *ptr = 0;                   /* terminate the string */
    return (seq);
}



/*++

Looks up key binding in binding table.  Returns pointer to
binding or NULL. Global helper fn.

++*/
KEYTAB *exist_binding(c)
int c;                          /* key */

{
    KEYTAB *ktp;

    ktp = get_binding(c);
    if (ktp->k_type != BINDNUL) {
        return (ktp);
    }

    return ((KEYTAB *) NULL);
}
